/**
 * bosfs - A fuse-based file system implemented on Baidu Object Storage(BOS)
 *
 * Copyright (c) 2016 Baidu.com, Inc. All rights reserved.
 *
 * @file    bosfs_util.cpp
 * @author  Song Shuangyang(songshuangyang@baidu.com)
 * @date    2016.9
 **/
#include "bosfs_util.h"
#include "file_manager.h"
#include <stdlib.h>    /* for malloc/free */
#include <unistd.h>    /* for sysconf, geteuid... */
#include <errno.h>     /* for error number */
#include <libgen.h>    /* for dirname/basename */
#include <pwd.h>      /* for passwd */
#include <grp.h>      /* for group */
#include <dirent.h>
#include <fstream>
#include <sstream>
#include <vector>
#include <pthread.h>
#include <fuse.h>

#include <cstdio>

BEGIN_FS_NAMESPACE

using namespace baidu::bos::cppsdk;

class FileGuard {
public:
    FileGuard() : _fp(NULL) {}
    ~FileGuard() {
        if (_fp != NULL) {
            fclose(_fp);
        }
    }

    void set(FILE *fp) {
        _fp = fp;
    }
private:
    FILE *_fp;
};
FileGuard g_log_file;

int g_reload_interval_s = 60;
bool g_reload_on = false;
pthread_mutex_t g_reload_mutex = PTHREAD_MUTEX_INITIALIZER;
pthread_cond_t g_reload_cond = PTHREAD_COND_INITIALIZER;

pthread_mutex_t g_client_mutex = PTHREAD_MUTEX_INITIALIZER;

SharedPtr<Client> BosfsUtil::_s_bos_client;

const char *        BosfsUtil::_s_default_mime_file = "/etc/mime.types";
BosfsUtil::mimes_t  BosfsUtil::_s_mime_types;


SharedPtr<Client> BosfsUtil::bos_client() {
    MutexGuard lock(&g_client_mutex);
    return _s_bos_client;
}

std::string get_real_path(const char*);
int BosfsUtil::reload() {
    LOG(INFO) << "reload credentials";
    int ret = BosfsUtil::load_credentials();
    if (ret != 0) {
        LOG(WARN) << "reload credentials failed: " << ret;
        return ret;
    }
    ret = BosfsUtil::create_bos_client();
    if (ret != 0) {
        LOG(WARN) << "create bos client failed: " << ret;
        return ret;
    }
    return 0;
}

std::string BosfsUtil::get_real_path(const char* path) {
    std::string realpath = "/" + g_bucket_prefix + std::string(path+1);
    if (*realpath.rbegin() == '/' && realpath != "/")
    {
        realpath = realpath.substr(0, realpath.size()-1); 
    }
    return realpath;
}
void BosfsUtil::reload_sighandler(int sig) {
    LOG(INFO) << "received signal " << sig << ", reloading credential: " << g_credentials_path;
    if (g_reload_on) {
        pthread_mutex_lock(&g_reload_mutex);
        pthread_cond_signal(&g_reload_cond);
        pthread_mutex_unlock(&g_reload_mutex);
    } else {
        BosfsUtil::reload();
    }
}

void *BosfsUtil::reload_thread(void *param) {
    (void) param;
    while (true) {
        int reload_interval_s = g_reload_interval_s;
        if (reload_interval_s <= 0) {
            reload_interval_s = 60;
        }
        timespec ts;
        clock_gettime(CLOCK_REALTIME, &ts);
        ts.tv_sec += reload_interval_s;

        MutexGuard lock(&g_reload_mutex);
        while (g_reload_on) {
            if (pthread_cond_timedwait(&g_reload_cond, &g_reload_mutex, &ts) != 0) {
                break;
            }
        }
        if (!g_reload_on) {
            break;
        }
        lock.unlock();
        reload();
    }
    return NULL;
}

void BosfsUtil::start_reload() {
    pthread_mutex_lock(&g_reload_mutex);
    if (!g_reload_on) {
        g_reload_on = true;
        pthread_attr_t attr;
        pthread_attr_init(&attr);
        pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
        pthread_t tid;
        pthread_create(&tid, &attr, reload_thread, NULL);
    }
    pthread_mutex_unlock(&g_reload_mutex);
}

void BosfsUtil::stop_reload() {
    pthread_mutex_lock(&g_reload_mutex);
    g_reload_on = false;
    pthread_cond_signal(&g_reload_cond);
    pthread_mutex_unlock(&g_reload_mutex);
}

void BosfsUtil::set_log_level(const std::string &level) {
    BOSFS_INFO("loglevel set to %s", level.c_str());
    int l = LogUtil::string_to_level(level.c_str());
    if (l >= 0) {
        sdk_set_log_level((bcesdk_ns::LogLevel) l);
    }
}

void BosfsUtil::set_log_file(const std::string &file) {
    FILE *fp = fopen(file.c_str(), "a");
    if (fp != NULL) {
        g_log_file.set(fp);
        sdk_set_log_stream(fp);
    }
}

int BosfsUtil::load_credentials() {
    g_ak = "";
    g_sk = "";
    g_sts_token = "";
    g_reload_interval_s = -1;

    std::ifstream f(g_credentials_path.c_str());
    if (f.fail()) {
        LOG(ERROR) << "open credentials " << g_credentials_path << " failed: (" << errno << ")"
            << strerror(errno);
        return -1;
    }
    std::string line;
    while (std::getline(f, line)) {
        line = StringUtil::trim(line);
        if (line.empty()) {
            continue;
        }
        if (line[0] == '[') {
            LOG(DEBUG) << "find conf file marker: " << line;
            continue;
        } else if (line[0] == '#') {
            LOG(DEBUG) << "find conf file comment: " << line;
            continue;
        }
        size_t pos = line.find('=');
        if (pos == std::string::npos) {
            LOG(DEBUG) << "find conf file non key-value line: " << line;
            continue;
        }
        std::string key = StringUtil::trim(line.substr(0, pos));
        std::string value = StringUtil::trim(line.substr(pos + 1));
        if (key == "bce_access_key_id") {
            g_ak = value;
        } else if (key == "bce_secret_access_key") {
            g_sk = value;
        } else if (key == "bce_security_token") {
            g_sts_token = value;
        } else if (key == "reload_interval_s") {
            g_reload_interval_s = atoi(value.c_str());
        } else {
            LOG(DEBUG) << "conf file ignored key-value: " << line;
        }
    }
    if (g_reload_interval_s > 0) {
        start_reload();
    }
    return 0;
}

int BosfsUtil::init_bos() {
    int ret = 0;
    // init ak/sk
    if (g_ak.empty() || g_sk.empty()) {
        ret = BosfsUtil::load_credentials();
        if (ret != 0) {
            return ret;
        }
    }
    // init client instance
    ret = BosfsUtil::create_bos_client();
    if (ret != 0) {
        return ret;
    }
    //check bucket_prefix exit
    bool is_dir_obj = false;
    bool is_prefix = false;
    std::string prefix = g_bucket_prefix;
    if (g_bucket_prefix != "") {
        if (*g_bucket_prefix.rbegin() == '/') {
           prefix = g_bucket_prefix.substr(0, g_bucket_prefix.size() - 1); 
        }
        ret = BosfsUtil::head_object(prefix, NULL, &is_dir_obj, &is_prefix);
        if (ret != 0) {
            return ret;
        }
        if (!is_dir_obj && !is_prefix) {
            BOSFS_ERR("not mounting a directory \n");
            printf("not mounting a directory, exit\n");
            return BOSFS_NOT_DIRECTORY;
        }
        return 0;
    }

    // check bucket existence
    ret = BosfsUtil::exist_bucket();
    if (ret != 0) {
        if (g_create_bucket) {
            ret = BosfsUtil::create_bucket();
        }
        if (ret != 0) {
            return ret;
        }
    }
    return 0;
}

void BosfsUtil::init_default_stat(struct stat *pst) {
    memset(pst, 0, sizeof(struct stat));
    pst->st_nlink = 1;
    pst->st_mode = g_mount_mode;
    pst->st_uid = g_is_bosfs_uid ? g_bosfs_uid : g_mount_uid;
    pst->st_gid = g_is_bosfs_gid ? g_bosfs_gid : g_mount_gid;
    pst->st_ctime = g_mount_time;
    pst->st_mtime = g_mount_time;
    pst->st_size = 0;
    pst->st_blocks = 0;
    pst->st_blksize = ST_BLKSIZE;
}

std::string BosfsUtil::object_to_path(const std::string &object){
    if (object.empty()) {
        return "";
    }
    std::ostringstream result;
    if (object[0] != '/') {
        result << "/";
    }
    if (*object.rbegin() == '/') {
        result << object.substr(0, object.length() - 1);
    } else {
        result << object;
    }
    return result.str();
}

std::string BosfsUtil::object_to_basename(const std::string &object, const std::string &prefix) {
    std::string result;
    if (*object.rbegin() == '/') {
        result = object.substr(prefix.length(), object.length() - prefix.length() - 1);
    } else {
        result = object.substr(prefix.length(), object.length() - prefix.length());
    }
    return result;
}

std::string BosfsUtil::get_username(uid_t uid)
{
    std::string ret("");
    static size_t max_len = 0;
    if (0 == max_len) {
        long res = sysconf(_SC_GETPW_R_SIZE_MAX);
        if (res < 0) {
            BOSFS_WARN("%s", "Could not get max pw length!");
            max_len = 0;
            return ret;
        }
        max_len = res;
    }

    char *buf = NULL;
    if (NULL == (buf = (char *)malloc(sizeof(char) * max_len))) {
        BOSFS_ERR("%s", "Failed to allocate memory!");
        free(buf);
        return ret;
    }

    struct passwd pwinfo;
    struct passwd *ppwinfo = NULL;
    if (0 != getpwuid_r(uid, &pwinfo, buf, max_len, &ppwinfo)) {
        BOSFS_WARN("%s", "Could not get pw information!");
        free(buf);
        return ret;
    }

    if (NULL == ppwinfo) {
        free(buf);
        return ret;
    }
    ret = ppwinfo->pw_name;
    free(buf);
    return ret;
}

int BosfsUtil::is_uid_in_group(uid_t uid, gid_t gid)
{
    int result = 0;
    static size_t max_len = 0;
    if (0 == max_len) {
        long res = sysconf(_SC_GETGR_R_SIZE_MAX);
        if (res < 0) {
            BOSFS_ERR("%s", "Could not get max name length!");
            max_len = 0;
            return -ERANGE;
        }
        max_len = res;
    }

    char * buf = NULL;
    if (NULL ==(buf = (char *)malloc(sizeof(char) * max_len))) {
        BOSFS_ERR("%s", "Failed to allocate memory!");
        free(buf);
        return -ENOMEM;
    }

    struct group ginfo;
    struct group *pginfo = NULL;
    if (0 != (result = getgrgid_r(gid, &ginfo, buf, max_len, &pginfo))) {
        BOSFS_ERR("%s", "Could not get group information!");
        free(buf);
        return -result;
    }

    if (NULL == pginfo) {
        free(buf);
        return -EINVAL;
    }

    std::string username = get_username(uid);
    char ** ppgr;
    for (ppgr = pginfo->gr_mem; ppgr && *ppgr; ++ppgr) {
        if (username == *ppgr) {
            free(buf);
            return 1;
        }
    }
    free(buf);
    return 0;
}

std::string BosfsUtil::bosfs_basename(const char *path)
{
    if (!path || '\0' == path[0]) {
        return std::string("");
    }
    return bosfs_basename(std::string(path));
}

std::string BosfsUtil::bosfs_basename(std::string path)
{
    return std::string(basename((char *)path.c_str()));
}

int BosfsUtil::mkdirp(const std::string &path, mode_t mode)
{
    std::string base;
    std::string component;
    std::stringstream ss(path);
    while (getline(ss, component, '/')) {
        base += "/" + component;

        struct stat st;
        if (0 == stat(base.c_str(), &st)) {
            if (!S_ISDIR(st.st_mode)) {
                return EPERM;
            }
        } else {
            if (0 != mkdir(base.c_str(), mode)) {
                return errno;
            }
        }
    }
    return 0;
}

bool BosfsUtil::check_exist_dir_permission(const char *dirpath)
{
    if (!dirpath || '\0' == dirpath[0]) {
        return false;
    }

    struct stat st;
    if (0 != stat(dirpath, &st)) {
        if (ENOENT == errno) {
            return true;
        }
        if (EACCES == errno) {
            return false;
        }
        return false;
    }

    if (!S_ISDIR(st.st_mode)) {
        return false;
    }

    uid_t uid = geteuid();
    if (uid == st.st_uid) {
        if (S_IRWXU != (st.st_mode & S_IRWXU)) {
            return false;
        }
    } else {
        if (1 == is_uid_in_group(uid, st.st_gid)) {
            if (S_IRWXG != (st.st_mode & S_IRWXG)) {
                return false;
            }
        } else {
            if (S_IRWXO != (st.st_mode & S_IRWXO)) {
                return false;
            }
        }
    }
    return true;
}

bool BosfsUtil::delete_files_in_dir(const char *dir, bool is_remove_own)
{
    DIR * pdir = NULL;
    struct dirent * dent;

    if (NULL == (pdir = opendir(dir))) {
        BOSFS_ERR("Could not open dir(%s) - errno(%d)", dir, errno);
        return false;
    }

    for (dent = readdir(pdir); dent; dent = readdir(pdir)) {
        if (0 == strcmp(dent->d_name, "..") || 0 == strcmp(dent->d_name, ".")) {
            continue;
        }

        std::string fullpath = dir;
        fullpath += "/";
        fullpath += dent->d_name;
        struct stat st;
        if (0 != lstat(fullpath.c_str(), &st)) {
            BOSFS_ERR("Could not open get stat of file(%s) - errno(%d)", fullpath.c_str(), errno);
            closedir(pdir);
            return false;
        }
        if (S_ISDIR(st.st_mode)) {
            if (!delete_files_in_dir(fullpath.c_str(), true)) {
                BOSFS_ERR("Could not remove sub dir(%s) - errno(%d)", fullpath.c_str(), errno);
                closedir(pdir);
                return false;
            }
        } else {
            if (0 != unlink(fullpath.c_str())) {
                BOSFS_ERR("Could not remove files(%s) - errno(%d)", fullpath.c_str(), errno);
                closedir(pdir);
                return false;
            }
        }
    }
    closedir(pdir);

    if (is_remove_own && 0 != rmdir(dir)) {
        BOSFS_ERR("Could not remove dir(%s) - errno(%d)", dir, errno);
        return false;
    }
    return true;
}

bool BosfsUtil::init_mimetype(const char *mime_file)
{
    if (!mime_file) {
        mime_file = _s_default_mime_file;
    }
    std::string line;
    std::ifstream ifs(mime_file);
    if (ifs.good()) {
        while (std::getline(ifs, line)) {
            if (line.size() == 0 || line[0] == '#') {
                continue;
            }

            std::stringstream tmp(line);
            std::string mimetype;
            tmp >> mimetype;
            while (tmp) {
                std::string ext;
                tmp >> ext;
                if (ext.size() == 0) {
                    continue;
                }
                _s_mime_types[ext] = mimetype;
            }
        }
    }
    return true;
}

std::string BosfsUtil::get_mimetype(const std::string &path_name)
{
    std::string result("application/octet-stream");
    std::string::size_type last_pos = path_name.find_last_of('.');
    std::string::size_type first_pos = path_name.find_first_of('.');
    std::string prefix;
    std::string ext;
    std::string ext2;

    if (last_pos == std::string::npos) {
        return result;
    } else {
        ext = path_name.substr(last_pos + 1, std::string::npos);

        if (first_pos != std::string::npos && first_pos < last_pos) {
            prefix = path_name.substr(0, last_pos);
            std::string::size_type next_pos = prefix.find_last_of('.');
            if (next_pos != std::string::npos) {
                ext2 = prefix.substr(next_pos + 1, std::string::npos);
            }
        }
    }

    // check the first ext
    mimes_t::const_iterator iter = _s_mime_types.find(ext);
    if (iter != _s_mime_types.end()) {
        result = iter->second;
        return result;
    }

    // check the second ext
    if (first_pos == last_pos) {
        return result;
    }
    iter = _s_mime_types.find(ext2);
    if (iter != _s_mime_types.end()) {
        result = iter->second;
    }
    return result;
}

bool BosfsUtil::set_mountpoint_attr(struct stat &ps)
{
    g_mount_uid = geteuid();
    g_mount_gid = getegid();
    g_mount_mode = S_IFDIR | (g_allow_other ? (~g_mount_umask & (S_IRWXU | S_IRWXG | S_IRWXO)) : S_IRWXU);
    BOSFS_INFO("PROC(uid=%u, gid=%u, mode=%04o) - Mountpoint(uid=%u, gid=%u, mode=%04o)",
            static_cast<unsigned int>(g_mount_uid), static_cast<unsigned int>(g_mount_gid),
            g_mount_mode,
            static_cast<unsigned int>(ps.st_uid), static_cast<unsigned int>(ps.st_gid),
            static_cast<unsigned int>(ps.st_mode));
    g_mount_time = time(NULL);

    if (0 == g_mount_uid || ps.st_uid == g_mount_uid) {
        return true;
    }

    if (ps.st_gid == g_mount_gid || 1 == is_uid_in_group(g_mount_uid, ps.st_gid)) {
        if (S_IRWXG == (ps.st_mode & S_IRWXG)) {
            return true;
        }
    }

    if (S_IRWXO == (ps.st_mode & S_IRWXO)) {
        return true;
    }
    return false;
}

int BosfsUtil::check_object_access(const char *path, int mask, struct stat *pstbuf)
{
    int ret = 0;

    struct fuse_context *pctx;
    if (NULL == (pctx = fuse_get_context())) {
        return -EIO;
    }

    struct stat st;
    struct stat *pst = pstbuf ? pstbuf : &st;
    ret = get_object_attribute(path, pst);
    if (ret != 0) {
        return ret;
    }

    if (0 == pctx->uid) { // root is allowed all accessing
        return 0;
    }

    if (g_is_bosfs_uid && g_bosfs_uid == pctx->uid) { // uid user is allowed
        return 0;
    }

    if (F_OK == mask) { // if there is a file return allowed
        return 0;
    }

    uid_t obj_uid = g_is_bosfs_uid ? g_bosfs_uid : pst->st_uid;
    gid_t obj_gid = g_is_bosfs_gid ? g_bosfs_gid : pst->st_gid;
    mode_t mode;
    mode_t base_mask = S_IRWXO;
    if (g_is_bosfs_umask) { // if umask is set, all object attr set ~umask
        mode = ((S_IRWXU | S_IRWXG | S_IRWXO) & ~g_bosfs_mask);
    } else {
        mode = pst->st_mode;
    }
    if (pctx->uid == obj_uid) {
        base_mask |= S_IRWXU;
    }
    if (pctx->gid == obj_gid) {
        base_mask |= S_IRWXG;
    }
    if (1 == is_uid_in_group(pctx->uid, obj_gid)) {
        base_mask |= S_IRWXG;
    }
    mode &= base_mask;

    if (X_OK == (mask & X_OK)) {
        if (0 == (mode & (S_IXUSR | S_IXGRP | S_IXOTH))) {
            return -EACCES;
        }
    }
    if (W_OK == (mask & W_OK)) {
        if (0 == (mode & (S_IWUSR | S_IWGRP | S_IWOTH))) {
            return -EACCES;
        }
    }
    if (R_OK == (mask & R_OK)) {
        if (0 == (mode & (S_IRUSR | S_IRGRP | S_IROTH))) {
            return -EACCES;
        }
    }
    if (0 == mode) {
        return -EACCES;
    }
    return 0;
}

int BosfsUtil::get_object_attribute(const std::string &path, struct stat *pstbuf,
        ObjectMetaData *pmeta) {
    struct stat tmpstbuf;
    struct stat *pst = pstbuf ? pstbuf : &tmpstbuf;
    init_default_stat(pst);
    if (path == "/" || path == ".") {
        pst->st_size = ST_BLKSIZE;
        pst->st_blocks = ST_MINBLOCKS;
        return 0;
    }

    FilePtr file;
    int ret = FileManager::instance().get(path, &file);
    if (ret != 0) {
        return ret;
    }
    if (pmeta != NULL) {
        *pmeta = file->meta();
    }
    if (pstbuf != NULL) {
        file->stat(pstbuf);
    }
    return 0;
}

// check each path component has searching permission
int BosfsUtil::check_path_accessible(const char *path) {
    std::string parent(path);
    size_t pos = parent.rfind('/');
    while (pos != 0) {
        parent.resize(pos);
        int ret = check_object_access(parent.c_str(), X_OK, NULL);
        if (ret != 0) {
            return ret;
        }
        pos = parent.rfind('/');
    }
    return 0;
}

int BosfsUtil::check_parent_object_access(const char *path, int mask) {
    int ret = 0;
    if (mask & X_OK) {
        ret = check_path_accessible(path);
        if (ret != 0) {
            return ret;
        }
    }
    mask = (mask & ~X_OK);
    if (mask) {
        std::string parent(path);
        size_t pos = parent.rfind('/');
        if (pos > 0) {
            parent.resize(pos);
        } else {
            parent.resize(1);
        }
        ret = check_object_access(parent.c_str(), mask, NULL);
    }
    return ret;
}

int BosfsUtil::check_object_owner(const char *path, struct stat *pstbuf)
{
    struct fuse_context *pctx = NULL;
    if (NULL == (pctx = fuse_get_context())) {
        return -EIO;
    }
    int ret = 0;
    struct stat st;
    struct stat *pst = (pstbuf ? pstbuf : &st);
    if (0 != (ret = get_object_attribute(path, pst))) {
        return ret;
    }

    if (0 == pctx->uid) {
        return 0;
    }

    if (g_is_bosfs_uid && g_bosfs_uid == pctx->uid) {
        return 0;
    }

    if (pctx->uid == pst->st_uid) {
        return 0;
    }
    return -EPERM;
}

DataCacheEntity * BosfsUtil::get_local_entity(const char *path, bool is_load)
{
    struct stat st;
    ObjectMetaData meta;
    if (0 != get_object_attribute(path, &st, &meta)) {
        return NULL;
    }

    time_t mtime = (!S_ISREG(st.st_mode) || S_ISLNK(st.st_mode)) ? -1 : st.st_mtime;
    bool force_tmpfile = S_ISREG(st.st_mode) ? false : true;

    DataCacheEntity *ent = NULL;
    if (NULL == (ent = DataCache::instance()->open_cache(path, &meta,
                    static_cast<ssize_t>(st.st_size), mtime, force_tmpfile, true))) {

        BOSFS_ERR("could not open file, errno = %d", errno);
        return NULL;
    }

    if (is_load && !ent->open_and_load_all(&meta)) {
        BOSFS_ERR("could not load file, errno = %d", errno);
        DataCache::instance()->close_cache(ent);
        return NULL;
    }

    return ent;
}


// Api relating to BOS client
int BosfsUtil::create_bos_client() {
    if (g_ak.empty() || g_sk.empty()) {
        BOSFS_ERR("initiate bos client error: %s", stringfy_ret_code(BOSFS_AK_SK_INVALID));
        return BOSFS_AK_SK_INVALID;
    }
    ClientOptions option;
    option.user_agent = std::string("bosfs-" VERSION "/") + sdk_package_string();
    if (g_endpoint.empty()) {
        option.endpoint = DEFAULT_ENDPOINT;
    } else {
        option.endpoint = g_endpoint;
    }
    if (g_bos_client_timeout <= 0) {
        BOSFS_ERR("initiate bos client error: %s", stringfy_ret_code(BOSFS_TIMEOUT_INVALID));
        return BOSFS_TIMEOUT_INVALID;
    }
    option.timeout = g_bos_client_timeout;
    option.multi_part_size = g_multipart_size;
    option.max_parallel = g_multipart_parallel;
    MutexGuard lock(&g_client_mutex);
    _s_bos_client.reset(new Client(Credential(g_ak, g_sk, g_sts_token), option));
    return BOSFS_OK;
}

int BosfsUtil::exist_bucket()
{
    BOSFS_INFO("check the bucket exist: %s", g_bucket.c_str());

    GetBucketLocationRequest request(g_bucket);
    GetBucketLocationResponse response;
    int ret = bos_client()->get_bucket_location(request, &response);

    if (0 != ret || response.is_fail()) {
        BOSFS_ERR("check bucket exist (%s) failed: %s, bos client errno: %d",
                g_bucket.c_str(), response.error().message().c_str(), ret);
        return BOSFS_BOS_CLIENT_REQUEST_ERROR;
    }
    if (g_endpoint.empty()) {
        g_endpoint = response.location() + ".bcebos.com";
        return create_bos_client();
    }
    return BOSFS_OK;
}

int BosfsUtil::create_bucket()
{
    BOSFS_INFO("create the bucket: %s", g_bucket.c_str());


    PutBucketRequest request(g_bucket);
    PutBucketResponse response;
    int ret = bos_client()->put_bucket(request, &response);

    if (0 != ret || response.is_fail()) {
        BOSFS_ERR("create bucket (%s) failed: %s, bos client errno: %d",
                g_bucket.c_str(), response.error().message().c_str(), ret);
        return BOSFS_CREATE_BUCKET_FAILED;
    }
    return BOSFS_OK;
}

int BosfsUtil::check_bucket_access()
{
    BOSFS_INFO("check the bucket access: %s", g_bucket.c_str());


    GetBucketAclRequest request(g_bucket);
    GetBucketAclResponse response;
    int ret = bos_client()->get_bucket_acl(request, &response);
    if (0 != ret || response.is_fail()) {
        BOSFS_ERR("check bucket access (%s) failed: %s, bos client errno: %d",
                g_bucket.c_str(), response.error().message().c_str(), ret);
        return BOSFS_BOS_CLIENT_REQUEST_ERROR;
    }

    bool can_read = false;
    bool can_write = false;
    std::string acl_id = response.owner().id;
    std::vector<Grant> acls = response.access_control_list();
    size_t acl_len = acls.size();
    for (size_t i = 0; i < acl_len; ++i) {
        std::vector<Grantee> &grantee = acls[i].grantee;
        std::vector<std::string> &perms = acls[i].permission;
        for (size_t j = 0; j < grantee.size(); ++j) {
            if ("*" == grantee[j].id || acl_id == grantee[j].id) {
                for (size_t k = 0; k < perms.size(); ++k) {
                    if ("READ" == perms[k]) {
                        can_read = true;
                    }
                    if ("WRITE" == perms[k]) {
                        can_write = true;
                    }
                    if ("FULL_CONTROL" == perms[k]) {
                        can_read = true;
                        can_write = true;
                    }
                }
            }
        }
    }

    if (can_read && can_write) {
        return BOSFS_OK;
    }
    return BOSFS_BUCKET_ACCESS_DENIED;
}

int BosfsUtil::head_object(const std::string &object, ObjectMetaData *meta, bool *is_dir_obj,
        bool *is_prefix) {
    BOSFS_INFO("head object request: %s/%s", g_bucket.c_str(), object.c_str());
    BceRequestContext ctx[2];
    HeadObjectRequest req(g_bucket, object);
    HeadObjectResponse res;
    HeadObjectRequest req1(g_bucket, object + "/");
    HeadObjectResponse res1;
    ctx[0].request = &req;
    ctx[0].response = &res;
    ctx[1].request = &req1;
    ctx[1].response = &res1;
    int ret = bos_client()->send_request(2, ctx);
    if (ret != 0) {
        return ret;
    }
    *is_dir_obj = false;
    *is_prefix = false;
    if (!res.is_fail()) {
        // regular file object
        if (meta != NULL) {
            meta->move_from(res.meta());
        }
        return BOSFS_OK;
    } else if (res.status_code() != 404) {
        BOSFS_WARN("head object(%s) failed, bos service error: %s", object.c_str(),
                res.error().message().c_str());
        return BOSFS_BOS_SERVICE_ERROR;
    }
    if (!res1.is_fail()) {
        // directory object
        *is_dir_obj = true;
        if (meta != NULL) {
            meta->move_from(res1.meta());
        } 
        return BOSFS_OK;
    } else if (res1.status_code() != 404) {
        BOSFS_WARN("head object(%s) failed, bos service error: %s", req1.object_name().c_str(),
                res1.error().message().c_str());
        return BOSFS_BOS_SERVICE_ERROR;
    }
    std::vector<std::string> subitems;
    if (BosfsUtil::list_subitems(object + "/", 2, &subitems) != 0) {
        return BOSFS_BOS_SERVICE_ERROR;
    }
    if (subitems.empty()) {
        return BOSFS_OBJECT_NOT_EXIST;
    }
    *is_prefix = true;
    return BOSFS_OK;
}

int BosfsUtil::multiple_head_object(std::vector<std::string> &objects,
        std::vector<struct stat *> &stats) {
    std::vector<BceRequestContext> ctx(objects.size());
    for (size_t i = 0; i < objects.size(); ++i) {
        ctx[i].request = new HeadObjectRequest(g_bucket, objects[i]);
        ctx[i].response = new HeadObjectResponse();
        ctx[i].is_own = true;
    }
    int ret = bos_client()->send_request(ctx.size(), &ctx.front(), 100);
    if (ret != 0) {
        return ret;
    }
    for (size_t i = 0; i < objects.size(); ++i) {
        HeadObjectResponse *res = (HeadObjectResponse *) ctx[i].response;
        init_default_stat(stats[i]);
        if (res->is_fail()) {
            if (!is_dir_path(objects[i])) {
                  stats[i]->st_mode &= ~S_IFDIR;
                  stats[i]->st_mode |= S_IFREG;
            }
            if (res->status_code() != 404) {
                LOG(WARN) << "get object " << objects[i] << "'s meta failed: "
                    << res->error().message() << ", return empty stat";
            }
        } else {
            bool is_dir_obj = false;
            std::string path = "/";
            if (*objects[i].rbegin() == '/') {
                path += objects[i].substr(0, objects[i].length() - 1);
                is_dir_obj = true;
            } else {
                path += objects[i];
            }
            FilePtr file(new File(path));
            file->meta().move_from(res->meta());
            file->set_is_dir_obj(is_dir_obj);
            file->stat(stats[i]);
            FileManager::instance().set(path, file);
        }
    }
    return 0;
}

int BosfsUtil::list_objects(const std::string &prefix, int max_keys, std::string &marker,
        const char *delimiter, std::vector<std::string> *items,
        std::vector<std::string> *common_prefix) {
    ListObjectsRequest request(g_bucket);
    if (max_keys >= 0 && max_keys < 1000) {
        request.set_max_keys(max_keys);
    }
    if (prefix == "" || is_dir_path(prefix)) {
        request.set_prefix(prefix);
    } else {
        request.set_prefix(prefix + "/");
    }
    if (delimiter != NULL) {
        request.set_delimiter(delimiter);
    }

    bool has_next = true;
    int n = 0;
    if (common_prefix == NULL) {
        common_prefix = items;
    }
    while (has_next && (max_keys <= 0 || n < max_keys)) {
        request.set_marker(marker);
        if (max_keys - n < 1000) {
            request.set_max_keys(max_keys - n);
        }
        ListObjectsResponse response;
        bos_client()->list_objects(request, &response);
        if (response.is_fail()) {
            LOG(ERROR) << "list objects [" << prefix << "] failed: (" << response.status_code()
                << ")" << response.error().message();
            return -1;
        }
        const std::vector<std::string> dirs = response.common_prefixes();
        for (size_t i = 0; i < dirs.size(); ++i) {
            common_prefix->push_back(dirs[i]);
        }
        n += dirs.size();

        const std::vector<ObjectSummary> &objects = response.contents();
        if (objects.size() > 0) {
            if (objects[0].key != prefix) {
                items->push_back(objects[0].key);
                n += objects.size();
            } else {
                n += objects.size() - 1;
            }
            for (size_t i = 1; i < objects.size(); ++i) {
                items->push_back(objects[i].key);
            }
        }
        marker = response.next_marker();
        has_next = response.is_truncated();
    }
    if (!has_next) {
        marker.clear();
    }
    return 0;
}

void BosfsUtil::create_meta(const std::string &object_name, mode_t mode, uid_t uid, gid_t gid,
        ObjectMetaData *meta) {
    if (S_ISLNK(mode)) {
        meta->set_content_type("application/octet-stream");
    } else if (S_ISDIR(mode)) {
        meta->set_content_type("application/x-directory");
    } else {
        meta->set_content_type(get_mimetype(object_name));
    }
    meta->set_user_meta("bosfs-uid", uid);
    meta->set_user_meta("bosfs-gid", gid);
    meta->set_user_meta("bosfs-mode", mode);
    meta->set_user_meta("bosfs-mtime", TimeUtil::now());
}

int BosfsUtil::create_object(const char *path, mode_t mode, uid_t uid, gid_t gid) {
    static const std::string emptystr;
    return create_object(path, mode, uid, gid, emptystr);
}

int BosfsUtil::create_object(const char *path, mode_t mode, uid_t uid, gid_t gid,
        const std::string &data) {
    BOSFS_INFO("[path=%s][mode=%04o]", path, mode);

    std::string object_name = path + 1;
    ObjectMetaData meta;
    create_meta(object_name, mode, uid, gid, &meta);
    
    if (S_ISDIR(mode)) {
        object_name += "/";
    }
    PutObjectRequest request = PutObjectRequest(g_bucket, object_name);
    if (!data.empty()) {
        request.set_data(data);
    }
    PutObjectResponse response;
    request.set_meta(&meta);
    int ret = bos_client()->put_object(request, &response);

    if (0 != ret || response.is_fail()) {
        BOSFS_ERR("create object(%s) failed: %s, bos client errno: %d",
                path, response.error().message().c_str(), ret);
        return BOSFS_BOS_CLIENT_REQUEST_ERROR;
    }
    return BOSFS_OK;
}

int BosfsUtil::delete_object(const std::string &object, std::string *version)
{
    BOSFS_INFO("delete object request: %s", object.c_str());

    DeleteObjectRequest request(g_bucket, object);
    DeleteObjectResponse response;
    int ret = bos_client()->delete_object(request, &response);
    if (0 != ret) {
        BOSFS_ERR("delete object(%s) failed, bos client errno: %d", object.c_str(), ret);
        return BOSFS_BOS_CLIENT_REQUEST_ERROR;
    }
    if (response.is_fail()) {
        if (response.status_code() != 404) {
            BOSFS_WARN("delete object(%s) failed, bos service error: %s", object.c_str(),
                    response.error().message().c_str());
            return BOSFS_BOS_SERVICE_ERROR;
        }
        return BOSFS_OBJECT_NOT_EXIST;
    }
    if (NULL != version) {
        *version = response.version();
    }
    return BOSFS_OK;
}

int BosfsUtil::change_object_meta(const std::string &object, ObjectMetaData &meta) {
    int ret = bos_client()->copy_object(g_bucket, object, g_bucket, object, "", &meta);
    if (ret == RET_KEY_NOT_EXIST) {
        //retry for 5 times, in case that object have not been flushed to bos
        for (int i = 0; i < 5; ++i) {
            ret = bos_client()->copy_object(g_bucket, object, g_bucket, object, "", &meta);
            if (ret == 0) {
                return 0; 
            }
            ::sleep(1);
        }
        return -ENOENT;
    } else if (ret != 0) {
        return -EIO;
    }
    return 0;
}

int BosfsUtil::rename_file(const std::string &src, const std::string &dst, int64_t size_hint) {
    BOSFS_INFO("copy object request from: %s to: %s", src.c_str(), dst.c_str());
    int ret = 0;
    if (size_hint < 0 || size_hint >= g_multipart_threshold) {
        ret = bos_client()->parallel_copy(g_bucket, src, g_bucket, dst, g_storage_class);
    } else {
        ret = bos_client()->copy_object(g_bucket, src, g_bucket, dst, g_storage_class);
    }
    if (ret != 0) {
        FileManager::instance().del("/" + src.substr(0, src.size()-1));
        FileManager::instance().del("/" + dst.substr(0, dst.size()-1));
        return ret;
    }
    delete_object(src);

    FileManager::instance().del("/" + src.substr(0, src.size()-1));
    FileManager::instance().del("/" + dst.substr(0, dst.size()-1));
    return 0;
}

int BosfsUtil::rename_directory(const std::string &src, const std::string &dst) {
    std::string prefix;
    std::string dst_prefix;
    if (*src.rbegin() == '/') {
        prefix = src;
    } else {
        prefix = src + '/';
    }
    if (*dst.rbegin() == '/') {
        dst_prefix = dst;
    } else {
        dst_prefix = dst + '/';
    }
    std::string marker;
    std::vector<std::string> items;
    int ret = list_objects(prefix, -1, marker, NULL, &items);
    if (ret != 0) {
        return ret;
    }
    std::vector<std::string> dst_items;
    for (size_t i = 0; i < items.size(); ++i) {
        std::string dst_object = dst_prefix + items[i].substr(prefix.length(),
                items[i].length() - prefix.length());
        ret = bos_client()->copy_object(g_bucket, items[i], g_bucket, dst_object, g_storage_class);
        if (ret != 0 && ret != RET_KEY_NOT_EXIST) {
            break;
        }
        dst_items.push_back(dst_object);
        FileManager::instance().del("/" + dst_object);
    }
    if (dst_items.size() != items.size()) {
        ret = RET_SERVICE_ERROR;
    } else {
        ret = rename_file(prefix, dst_prefix, 0);
    }
    if (ret != 0 && ret != RET_KEY_NOT_EXIST) {
        for (size_t i = 0; i < dst_items.size(); ++i) {
            delete_object(dst_items[i]);
            FileManager::instance().del("/" + dst_items[i]);
        }
        return BOSFS_BOS_SERVICE_ERROR;
    }
    for (size_t i = 0; i < items.size(); ++i) {
        delete_object(items[i]);
        FileManager::instance().del("/" + items[i]);
    }
    return 0;
}

END_FS_NAMESPACE

